@@ -1,8 +1,11 @@
+from typing import Optional
+
 from .t import Direction, In
 from .digital import DigitalMeta, Digital
 from .wire import wire
 from magma.bit import Bit
 from magma.array import Array
+from magma.debug import debug_wire
 from magma.tuple import Tuple
 
 
@@ -15,7 +18,16 @@ class _ClockType(Digital):
         Bit.unused(self)
 
     def undriven(self):
-        Bit.unused(self)
+        Bit.undriven(self)
+
+    @classmethod
+    def is_wireable(cls, other):
+        # Wiring requires strict subclasses
+        # Note: we use the standard wiring logic to enforce directionality,
+        # so we just check with the undirected type here
+        if not issubclass(other, cls.qualify(Direction.Undirected)):
+            return False
+        return True
 
 
 class Clock(_ClockType, metaclass=DigitalMeta):
@@ -26,8 +38,12 @@ ClockIn = Clock[Direction.In]
 ClockOut = Clock[Direction.Out]
 
 
+class AbstractReset(_ClockType, metaclass=DigitalMeta):
+    pass
+
+
 # synchronous reset, active high (i.e. reset when signal is 1)
-class Reset(_ClockType, metaclass=DigitalMeta):
+class Reset(AbstractReset):
     pass
 
 
@@ -36,7 +52,7 @@ ResetOut = Reset[Direction.Out]
 
 
 # synchronous reset, active low (i.e. reset when signal is 0)
-class ResetN(_ClockType, metaclass=DigitalMeta):
+class ResetN(AbstractReset):
     pass
 
 
@@ -45,7 +61,7 @@ ResetNOut = ResetN[Direction.Out]
 
 
 # asynchronous reset, active high (i.e. reset when signal is 1)
-class AsyncReset(_ClockType, metaclass=DigitalMeta):
+class AsyncReset(AbstractReset):
     pass
 
 
@@ -54,7 +70,7 @@ AsyncResetOut = AsyncReset[Direction.Out]
 
 
 # asynchronous reset, active low (i.e. reset when signal is 0)
-class AsyncResetN(_ClockType, metaclass=DigitalMeta):
+class AsyncResetN(AbstractReset):
     pass
 
 
@@ -64,14 +80,14 @@ AsyncResetNOut = AsyncResetN[Direction.Out]
 
 # Preset
 # Clear
-class Enable(_ClockType, metaclass=DigitalMeta):
+class Enable(Bit):
     pass
 
 
 EnableIn = Enable[Direction.In]
 EnableOut = Enable[Direction.Out]
 
-ClockTypes = (Clock, Reset, AsyncReset, Enable)
+ClockTypes = (Clock, Reset, ResetN, AsyncReset, AsyncResetN, Enable)
 
 
 def ClockInterface(has_enable=False, has_reset=False, has_ce=False,
@@ -89,39 +105,28 @@ def ClockInterface(has_enable=False, has_reset=False, has_ce=False,
     return args
 
 
-def _wire_clock_port(port, clocktype, defnclk):
-    if isinstance(port, (Array, Tuple)):
-        for elem in port:
-            _wire_clock_port(elem, clocktype, defnclk)
-        return
-    if isinstance(port, clocktype) and not port.driven():
-        wire(defnclk, port)
+def get_reset_args(reset_type: Optional[AbstractReset]):
+    if reset_type is None:
+        return tuple(False for _ in range(4))
+    if not issubclass(reset_type, AbstractReset):
+        raise TypeError(
+            f"Expected subclass of AbstractReset for argument reset_type, "
+            f"not {type(reset_type)}")
+
+    has_async_reset = issubclass(reset_type, AsyncReset)
+    has_async_resetn = issubclass(reset_type, AsyncResetN)
+    has_reset = issubclass(reset_type, Reset)
+    has_resetn = issubclass(reset_type, ResetN)
+    return (has_async_reset, has_async_resetn, has_reset, has_resetn)
 
 
-def _get_clocks(port, clocktype):
-    if isinstance(port, (Array, Tuple)):
-        return sum([_get_clocks(elem, clocktype) for elem in port], [])
-    if isinstance(port, clocktype) and port.is_output():
-        return [port]
-    return []
-
-
-def wireclocktype(defn, inst, clocktype):
-    defnclk = []
-    for port in defn.interface.ports.values():
-        defnclk += _get_clocks(port, clocktype)
-    if defnclk:
-        defnclk = defnclk[0]  # wire first clock
-        for port in inst.interface.inputs(include_clocks=True):
-            _wire_clock_port(port, clocktype, defnclk)
-
-
-def wiredefaultclock(defn, inst):
-    wireclocktype(defn, inst, Clock)
-
-
-def wireclock(define, circuit):
-    wireclocktype(define, circuit, Reset)
-    wireclocktype(define, circuit, AsyncReset)
-    wireclocktype(define, circuit, AsyncResetN)
-    wireclocktype(define, circuit, Enable)
+def is_clock_or_nested_clock(p, types=ClockTypes):
+    if issubclass(p, types):
+        return True
+    if issubclass(p, Array):
+        return is_clock_or_nested_clock(p.T, types)
+    if issubclass(p, Tuple):
+        for item in p.types():
+            if is_clock_or_nested_clock(item, types):
+                return True
+    return False
