@@ -1,11 +1,15 @@
 from functools import lru_cache
 import weakref
 from abc import ABCMeta
+import hwtypes as ht
 from .t import Kind, Direction, Type, In, Out
 from .debug import debug_wire, get_callee_frame_info
 from .compatibility import IntegerTypes
 from .logging import root_logger
-from magma.wire_container import Wire
+from .protocol_type import magma_type, magma_value
+
+from magma.ref import ArrayRef
+from magma.wire_container import Wire, WiringLog, Wireable
 
 
 _logger = root_logger()
@@ -79,11 +83,12 @@ class DigitalMeta(ABCMeta, Kind):
 
     def __call__(cls, value=None, *args, **kwargs):
         if value is not None:
-            if isinstance(value, (bool, IntegerTypes)):
+            if isinstance(value, (bool, IntegerTypes, ht.Bit)):
                 return cls.VCC if value else cls.GND
         result = super().__call__(*args, **kwargs)
         if value is not None:
-            assert isinstance(value, Digital), type(value)
+            if not isinstance(value, Digital):
+                raise TypeError(type(value))
             result(value)
         return result
 
@@ -119,21 +124,36 @@ class DigitalMeta(ABCMeta, Kind):
     def __len__(cls):
         return 1
 
+    @lru_cache()
     def qualify(cls, direction):
         return cls[direction]
 
     def __eq__(cls, rhs):
-        return cls is rhs
+        return isinstance(rhs, DigitalMeta)
+
+    @lru_cache()
+    def is_wireable(cls, rhs):
+        if issubclass(rhs, (bool, int, ht.Bit)):
+            return True
+        rhs = magma_type(rhs)
+        # allows undirected types to match (e.g. for temporary values)
+        return (issubclass(rhs.flip(), cls) or issubclass(cls, rhs.flip()) or
+                cls.qualify(Direction.Undirected) is rhs or
+                rhs.qualify(Direction.Undirected) is cls)
+
+    def is_bindable(cls, rhs):
+        return issubclass(cls, magma_type(rhs))
 
     __hash__ = type.__hash__
 
 
-class Digital(Type, metaclass=DigitalMeta):
+class Digital(Type, Wireable, metaclass=DigitalMeta):
     def __init__(self, value=None, name=None):
-        super().__init__(name=name)
-        self._wire = Wire(self)
+        Type.__init__(self, name=name)
+        Wireable.__init__(self)
 
     @classmethod
+    @lru_cache()
     def is_oriented(cls, direction):
         if not cls.is_directed:
             return direction == Direction.Undirected
@@ -148,41 +168,28 @@ class Digital(Type, metaclass=DigitalMeta):
 
     @debug_wire
     def wire(self, o, debug_info):
-        i = self
         # promote integer types to LOW/HIGH
-        if isinstance(o, IntegerTypes):
+        if isinstance(o, (IntegerTypes, bool, ht.Bit)):
             o = HIGH if o else LOW
 
+        o = magma_value(o)
         if not isinstance(o, Digital):
-            _logger.error(f'Cannot wire {i.debug_name} (type={type(i)}) to {o} '
-                          f'(type={type(o)}) because {o.debug_name} is not a '
-                          f'Digital', debug_info=debug_info)
+            _logger.error(
+                WiringLog(f"Cannot wire {{}} (type={type(self)}) to {o} "
+                          f"(type={type(o)}) because {{}} is not a Digital",
+                          self, o),
+                debug_info=debug_info
+            )
             return
+        Wireable.wire(self, o, debug_info)
 
-        i._wire.connect(o._wire, debug_info)
-        i.debug_info = debug_info
-        o.debug_info = debug_info
+    @debug_wire
+    def unwire(self, o=None, debug_info=None):
+        return Wireable.unwire(self, o, debug_info)
 
     def iswhole(self):
         return True
 
-    def wired(self):
-        return self._wire.wired()
-
-    # return the input or output Bit connected to this Bit
-    def trace(self):
-        return self._wire.trace()
-
-    # return the output Bit connected to this input Bit
-    def value(self):
-        return self._wire.value()
-
-    def driven(self):
-        return self._wire.driven()
-
-    def driving(self):
-        return self._wire.driving()
-
     @classmethod
     def unflatten(cls, value):
         if len(value) != 1 or not isinstance(value[0], Digital):
@@ -196,9 +203,6 @@ class Digital(Type, metaclass=DigitalMeta):
         cls = type(self)
         return self is cls.VCC or self is cls.GND
 
-    def unwire(i, o):
-        i._wire.unwire(o._wire)
-
     @classmethod
     def flat_length(cls):
         return 1
@@ -223,6 +227,27 @@ class Digital(Type, metaclass=DigitalMeta):
             return "GND"
         return Type.__repr__(self)
 
+    @property
+    def debug_name(self):
+        if self.const():
+            return repr(self)
+        return super().debug_name
+
+    def __bool__(self) -> bool:
+        if not self.const():
+            raise ValueError(
+                "Converting non-constant magma bit to bool not supported")
+        if self is type(self).VCC:
+            return True
+        assert self is type(self).GND
+        return False
+
+    def __int__(self) -> int:
+        return int(bool(self))
+
+    def has_children(self):
+        return False
+
 
 VCC = Digital.VCC
 GND = Digital.GND
