@@ -1,11 +1,11 @@
 from abc import ABC, abstractmethod
-from collections import OrderedDict
 from itertools import chain
-from .common import deprecated, setattrs
+from .common import deprecated
 from .compatibility import IntegerTypes, StringTypes
 from .protocol_type import MagmaProtocolMeta
 from .ref import InstRef, DefnRef, LazyDefnRef, LazyInstRef, NamedRef
 from .t import Type, Kind, Direction
+import weakref
 
 
 __all__  = ['make_interface']
@@ -19,7 +19,7 @@ def _flatten(l):
 
 
 def _make_interface_name(decl):
-    return f"Interface({', '.join([str(d) for d in decl])})"
+    return f"Interface"
 
 
 def _is_valid_port(port):
@@ -84,12 +84,12 @@ def _make_wire_str(driver, value, wired):
     if (value, driver) in wired:
         return ""
     wired.add((value, driver))
-    if driver.iswhole():
-        iname = value.name.qualifiedname()
-        oname = driver.name.qualifiedname()
-        return f"wire({oname}, {iname})\n"
-    return "".join(_make_wire_str(d, v, wired)
-                   for d, v in zip(driver, value))
+    if driver.has_children() and driver.name.anon():
+        return "".join(_make_wire_str(d, v, wired)
+                       for d, v in zip(driver, value))
+    iname = value.name.qualifiedname()
+    oname = driver.name.qualifiedname()
+    return f"wire({oname}, {iname})\n"
 
 
 def _make_wires(value, wired):
@@ -111,31 +111,36 @@ def _make_wires(value, wired):
     if value.is_output():
         return ""
     if value.is_mixed():
-        return "".join(_make_wires(v, wired) for v in value)
+        return "".join(_make_wires(v, wired)
+                       for v, _ in value.connection_iter())
     driver = value.value()
+    while driver is not None and driver.is_driven_anon_temporary():
+        driver = driver.value()
     if driver is None:
         return ""
-    if not driver.iswhole():
-        return "".join(_make_wires(v, wired) for v in value)
-    while driver is not None and driver.name.anon() and not driver.is_output():
-        driver = driver.value()  # skip anon values
+    if driver.has_children() and driver.name.anon():
+        return "".join(_make_wires(v, wired)
+                       for v, _ in value.connection_iter())
     s = ""
     while driver is not None:
         if (driver, value) in wired:
             break
         s += _make_wire_str(driver, value, wired)
-        if driver.is_output():
+        if not driver.driven():
             break
         value = driver
         driver = driver.value()
     return s
 
 
+def _dict_from_decl(decl):
+    return dict(zip(decl[::2], decl[1::2]))
+
+
 class InterfaceKind(Kind):
     def __init__(cls, *args, **kwargs):
         super().__init__(*args, **kwargs)
-        args = zip(cls._decl[::2], cls._decl[1::2])
-        cls.ports = OrderedDict(args)
+        cls.ports = _dict_from_decl(cls._decl)
 
     def items(cls):
         return cls.ports.items()
@@ -143,12 +148,20 @@ class InterfaceKind(Kind):
     def __iter__(cls):
         return iter(cls.ports)
 
-    def __str__(cls):
+    def args_to_str(cls):
         args = [f"\"{arg}\"" if i % 2 == 0 else str(arg)
                 for i, arg in enumerate(cls._decl)]
         return ", ".join(args)
 
+    def __str__(cls):
+        return f"Interface({cls.args_to_str()})"
+
+    def __repr__(cls):
+        return str(cls)
+
     def __eq__(cls, rhs):
+        if not isinstance(rhs, InterfaceKind):
+            return False
         return cls._decl == rhs._decl
 
     def __ne__(cls, rhs):
@@ -274,7 +287,7 @@ class Interface(_InterfaceBase):
 
         names, types = _parse_decl(decl)
         names = map(_map_name, names)
-        self.ports = OrderedDict(zip(names, types))
+        self.ports = dict(zip(names, types))
 
     def __str__(self):
         s = ", ".join(f"{k}: {v}" for k, v in self.ports.items())
@@ -321,6 +334,8 @@ class _DeclareInterface(_InterfaceBase):
         args = zip(names, zip(types, refs))
         flip = defn is not None
         self.ports = {n: _make_port(t, r, flip) for n, (t, r) in args}
+        for port in self.ports.values():
+            port.name._value = weakref.ref(port)
         _rename_ports(self.ports, renamed_ports)
 
 
@@ -336,7 +351,7 @@ class _DeclareSingletonInterface(_DeclareInterface):
             return
         assert defn is not None
         cls._io.bind(defn)  # bind IO to @defn
-        self.ports = OrderedDict(cls._io.ports.items())
+        self.ports = dict(cls._io.ports.items())
         _rename_ports(self.ports, renamed_ports)  # rename ports
         cls._initialized = True
 
@@ -354,7 +369,7 @@ class _DeclareSingletonInstanceInterface(_DeclareInterface):
                 super().__init__(renamed_ports, inst=inst, defn=defn)
                 return
             cls._io.bind(defn)  # bind IO to @defn
-            self.ports = OrderedDict(cls._io.ports.items())
+            self.ports = dict(cls._io.ports.items())
             _rename_ports(self.ports, renamed_ports)  # rename ports
             cls._initialized = True
             return
@@ -363,7 +378,7 @@ class _DeclareSingletonInstanceInterface(_DeclareInterface):
                 super().__init__(renamed_ports, inst=inst, defn=defn)
                 return
             cls._io.bind_inst(inst)  # bind IO to @inst
-            self.ports = OrderedDict(cls._io.inst_ports.items())
+            self.ports = dict(cls._io.inst_ports.items())
             _rename_ports(self.ports, renamed_ports)  # rename ports
             cls._initialized_inst = True
             return
@@ -410,6 +425,9 @@ class IOInterface(ABC):
         # __iadd__ is explicitly overriden to enforce that it is non-mutating.
         return self + other
 
+    def flip(self) -> "IOInterface":
+        raise NotImplementedError()
+
 
 class IO(IOInterface):
     """
@@ -423,14 +441,10 @@ class IO(IOInterface):
     def __init__(self, **kwargs):
         self._ports = {}
         self._decl = []
-        names = kwargs.keys()
-        types = kwargs.values()
-        refs = (LazyDefnRef(name=name) for name in kwargs)
-        args = zip(kwargs.keys(), zip(kwargs.values(), refs))
-        self._ports = {n: _make_port(t, r, flip=True) for n, (t, r) in args}
-        setattrs(self, self._ports)
-        self._decl = _flatten(zip(names, types))
         self._bound = False
+        for name, typ in kwargs.items():
+            self.add(name, typ)
+            self._decl.extend((name, typ))
 
     @property
     def ports(self):
@@ -471,7 +485,34 @@ class IO(IOInterface):
         if self._ports.keys() & other._ports.keys():
             raise Exception("Adding IO with duplicate port names not allowed")
         decl = self._decl + other._decl
-        return IO(**dict(zip(decl[::2], decl[1::2])))
+        return IO(**_dict_from_decl(decl))
+
+    def add(self, name, typ):
+        if self._bound:
+            raise RuntimeError("Can not add to a bound IO")
+        # Definition port.
+        ref = LazyDefnRef(name=name)
+        port = _make_port(typ, ref, flip=True)
+        self._ports[name] = port
+
+    def _get(self, key: str):
+        return self._ports[key]
+
+    def __getitem__(self, key: str):
+        return self._get(key)
+
+    def __getattr__(self, key: str):
+        try:
+            return self._get(key)
+        except KeyError:
+            pass
+        return super().__getattribute__(key)
+
+    def fields(self):
+        return _dict_from_decl(self._decl)
+
+    def flip(self):
+        return IO(**{name: T.flip() for name, T in self.fields().items()})
 
 
 class SingletonInstanceIO(IO):
@@ -496,6 +537,10 @@ class SingletonInstanceIO(IO):
     def inst_ports(self):
         return self._inst_ports.copy()
 
+    def decl(self):
+        return _flatten((name, type(port))
+                        for name, port in self._ports.items())
+
     def make_interface(self):
         decl = self.decl()
         name = _make_interface_name(decl)
@@ -504,11 +549,7 @@ class SingletonInstanceIO(IO):
         return InterfaceKind(name, (_DeclareSingletonInstanceInterface,), dct)
 
     def add(self, name, typ):
-        # Definition port.
-        ref = LazyDefnRef(name=name)
-        port = _make_port(typ, ref, flip=True)
-        self._ports[name] = port
-        setattr(self, name, port)
+        super().add(name, typ)
         # Instance port.
         inst_ref = LazyInstRef(name=name)
         inst_port = _make_port(typ, inst_ref, flip=False)
@@ -517,3 +558,6 @@ class SingletonInstanceIO(IO):
     def __add__(self, other):
         raise NotImplementedError(f"Addition operator disallowed on "
                                   f"{cls.__name__}")
+
+    def flip(self):
+        raise NotImplementedError()
