@@ -1,6 +1,17 @@
 import itertools
+from functools import lru_cache
+import functools
+import operator
 from collections import OrderedDict
-from hwtypes.adt import TupleMeta, Tuple as Tuple_, Product, ProductMeta
+from hwtypes.adt import (
+    TupleMeta,
+    Tuple as Tuple_,
+    AnonymousProduct,
+    AnonymousProductMeta,
+    Product,
+    ProductMeta,
+)
+from hwtypes import BitVector, Bit
 from hwtypes.adt_meta import BoundMeta, RESERVED_SUNDERS
 from hwtypes.util import TypedProperty, OrderedFrozenDict
 from .common import deprecated
@@ -9,6 +20,12 @@ from .t import Type, Kind, Direction
 from .compatibility import IntegerTypes
 from .debug import debug_wire, get_callee_frame_info
 from .logging import root_logger
+from .protocol_type import magma_type, magma_value
+
+from magma.wire_container import WiringLog
+from magma.wire import wire
+from magma.protocol_type import MagmaProtocol
+from magma.operator_utils import output_only
 
 
 _logger = root_logger()
@@ -38,9 +55,22 @@ class TupleKind(TupleMeta, Kind):
                     has_bound_base = True
                     if bound_types is None:
                         bound_types = base.fields
-                    elif any(not (issubclass(x, y) or issubclass(y, x)) for x,
-                             y in zip(bound_types, base.fields)):
-                        raise TypeError("Can't inherit from multiple different bound_types")
+                    else:
+                        for x, y in zip(bound_types, base.fields):
+                            if not (issubclass(x, y) or issubclass(y, x)):
+                                try:
+                                    if x.is_bindable(y):
+                                        continue
+                                except AttributeError:
+                                    pass
+                                try:
+                                    if y.is_bindable(x):
+                                        continue
+                                except AttributeError:
+                                    pass
+                                raise TypeError(
+                                    "Can't inherit from multiple different bound_types"
+                                )
                 else:
                     unbound_bases.append(base)
 
@@ -73,7 +103,7 @@ class TupleKind(TupleMeta, Kind):
             raise TypeError('Type is already bound')
 
         undirected_idx = tuple(v.qualify(Direction.Undirected) for v in idx)
-        if undirected_idx != idx:
+        if any(x is not y for x, y in zip(undirected_idx, idx)):
             bases = [cls[undirected_idx]]
         else:
             bases = [cls]
@@ -99,6 +129,15 @@ class TupleKind(TupleMeta, Kind):
             new_fields.append(T.flip())
         return cls.unbound_t[new_fields]
 
+    @property
+    @lru_cache()
+    def direction(cls):
+        directions = iter(t.direction for t in cls.fields)
+        first = next(directions)
+        if all(d == first for d in directions):
+            return first
+        return None
+
     def __call__(cls, *args, **kwargs):
         obj = cls.__new__(cls, *args)
         obj.__init__(*args, **kwargs)
@@ -106,7 +145,47 @@ class TupleKind(TupleMeta, Kind):
 
     @property
     def N(cls):
-        return len(cls)
+        return len(cls.fields)
+
+    def __len__(cls):
+        return cls.N
+
+    def __str__(cls):
+        if not cls.is_bound:
+            return cls.__name__
+        s = "Tuple("
+        s += ",".join(str(T) for T in cls.fields)
+        s += ")"
+        return s
+
+    def is_wireable(cls, rhs):
+        rhs = magma_type(rhs)
+        if not isinstance(rhs, TupleKind) or len(cls.fields) != len(rhs.fields):
+            return False
+        for idx, T in enumerate(cls.fields):
+            if not T.is_wireable(rhs[idx]):
+                return False
+        return True
+
+    def is_bindable(cls, rhs):
+        rhs = magma_type(rhs)
+        if not isinstance(rhs, TupleKind) or len(cls.fields) != len(rhs.fields):
+            return False
+        for idx, T in enumerate(cls.fields):
+            if not T.is_bindable(rhs[idx]):
+                return False
+        return True
+
+    def __eq__(cls, rhs):
+        if not isinstance(rhs, TupleKind):
+            return False
+
+        if not cls.is_bound:
+            return not rhs.is_bound
+
+        return cls.fields == rhs.fields
+
+    __hash__ = TupleMeta.__hash__
 
 
 class Tuple(Type, Tuple_, metaclass=TupleKind):
@@ -118,16 +197,20 @@ class Tuple(Type, Tuple_, metaclass=TupleKind):
         if len(largs) > 0:
             assert len(largs) == len(self)
             for k, t, T in zip(self.keys(), largs, self.types()):
-                if type(t) is bool:
+                if isinstance(t, (IntegerTypes, BitVector, Bit)):
                     t = T(t)
                 self.ts.append(t)
-                if not isinstance(self, Product):
+                if not isinstance(self, AnonProduct):
                     setattr(self, k, t)
         else:
             for k, T in zip(self.keys(), self.types()):
-                t = T(name=TupleRef(self, k))
+                ref = TupleRef(self, k)
+                if issubclass(T, MagmaProtocol):
+                    t = T._from_magma_value_(T._to_magma_()(name=ref))
+                else:
+                    t = T(name=ref)
                 self.ts.append(t)
-                if not isinstance(self, Product):
+                if not isinstance(self, AnonProduct):
                     setattr(self, k, t)
 
     __hash__ = Type.__hash__
@@ -149,11 +232,19 @@ class Tuple(Type, Tuple_, metaclass=TupleKind):
     def keys(cls):
         return [str(i) for i in range(len(cls.fields))]
 
+    @output_only("Cannot use == on an input")
     def __eq__(self, rhs):
         if not isinstance(rhs, type(self)):
             return NotImplemented
         else:
-            return self.ts == rhs.ts
+            # NOTE: Use functools.reduce so import * doesn't clobber m.reduce
+            # from bits
+            return functools.reduce(operator.and_,
+                                    (x == y for x, y in zip(self, rhs)))
+
+    @output_only("Cannot use != on an input")
+    def __ne__(self, rhs):
+        return ~(self == rhs)
 
     def __repr__(self):
         if not self.name.anon():
@@ -170,15 +261,17 @@ class Tuple(Type, Tuple_, metaclass=TupleKind):
     def __setitem__(self, key, val):
         old = self[key]
         if old is not val:
-            _logger.error(f'May not mutate Tuple, trying to replace '
-                          f'{self}[{key}] ({old}) with {val}')
+            _logger.error(
+                WiringLog(f"May not mutate tuple, trying to replace "
+                          f"{{}}[{key}] ({{}}) with {{}}", self, old, val)
+            )
 
     def __len__(self):
         return len(type(self).fields)
 
     @classmethod
     def flat_length(cls):
-        return sum(T.flat_length() for T in cls.types())
+        return sum(magma_type(T).flat_length() for T in cls.types())
 
     def __call__(self, o):
         return self.wire(o, get_callee_frame_info())
@@ -186,32 +279,40 @@ class Tuple(Type, Tuple_, metaclass=TupleKind):
 
     @debug_wire
     def wire(i, o, debug_info):
-        # print('Tuple.wire(', o, ', ', i, ')')
-
+        o = magma_value(o)
         if not isinstance(o, Tuple):
-            _logger.error(f'Cannot wire {o.debug_name} (type={type(o)}) to {i.debug_name} (type={type(i)}) because {o.debug_name} is not a Tuple', debug_info=debug_info)  # noqa
+            _logger.error(
+                WiringLog(f"Cannot wire {{}} (type={type(o)}) to {{}} "
+                          f"(type={type(i)}) because {{}} is not a Tuple",
+                          o, i, o),
+                debug_info=debug_info
+            )
             return
 
         if i.keys() != o.keys():
-            _logger.error(f'Cannot wire {o.debug_name} (type={type(o)}, keys={list(i.keys())}) to {i.debug_name} (type={type(i)}, keys={list(o.keys())}) because the tuples do not have the same keys', debug_info=debug_info)  # noqa
+            _logger.error(
+                WiringLog(f"Cannot wire {{}} (type={type(o)}, "
+                          f"keys={list(i.keys())}) to {{}} (type={type(i)}, "
+                          f"keys={list(o.keys())}) because the tuples do not "
+                          f"have the same keys", o, i),
+                debug_info=debug_info
+            )
             return
 
-        #if i.Ts != o.Ts:
-        #    print('Wiring error: Tuple elements must have the same type')
-        #    return
-
         for i_elem, o_elem in zip(i, o):
-            if o_elem.is_input():
-                o_elem.wire(i_elem, debug_info)
-            else:
-                i_elem.wire(o_elem, debug_info)
+            i_elem = magma_value(i_elem)
+            o_elem = magma_value(o_elem)
+            wire(o_elem, i_elem, debug_info)
 
-    def unwire(i, o):
-        for i_elem, o_elem in zip(i, o):
-            if o_elem.is_input():
-                o_elem.unwire(i_elem)
+    @debug_wire
+    def unwire(self, o=None, debug_info=None):
+        for k, t in self.items():
+            if o is None:
+                t.unwire(debug_info=debug_info)
+            elif o[k].is_input():
+                o[k].unwire(t, debug_info=debug_info)
             else:
-                i_elem.unwire(o_elem)
+                t.unwire(o[k], debug_info=debug_info)
 
     def driven(self):
         for t in self.ts:
@@ -259,12 +360,15 @@ class Tuple(Type, Tuple_, metaclass=TupleKind):
     def iswhole(self):
         return Tuple._iswhole(self.ts, self.keys())
 
-
-    def trace(self):
-        ts = [t.trace() for t in self.ts]
-
-        for t in ts:
-            if t is None:
+    def trace(self, skip_self=True):
+        ts = []
+        for t in self.ts:
+            result = t.trace(skip_self)
+            if result is not None:
+                ts.append(result)
+            elif not skip_self and (t.is_output() or t.is_inout()):
+                ts.append(t)
+            else:
                 return None
 
         if len(ts) == len(self) and Tuple._iswhole(ts, self.keys()):
@@ -336,55 +440,73 @@ class Tuple(Type, Tuple_, metaclass=TupleKind):
             mixed |= field.is_mixed()
         return mixed or (input + output + inout) > 1
 
-class ProductKind(ProductMeta, TupleKind, Kind):
+    def connection_iter(self, only_slice_bits=False):
+        for elem in self:
+            yield elem, elem.trace()
+
+    def has_children(self):
+        return True
+
+    def set_when_context(self, ctx):
+        for value in self:
+            value.set_when_context(ctx)
+
+
+def _add_properties(ns, fields):
+    def _make_prop(field_type, idx):
+        @TypedProperty(field_type)
+        def prop(self):
+            return self[idx]
+
+        @prop.setter
+        def prop(self, value):
+            self[idx] = value
+
+        return prop
+
+    for idx, (field_name, field_type) in enumerate(fields.items()):
+        if field_name == "N":
+            # TODO: Make N unreserved
+            raise ValueError("N is a reserved name in Product")
+        assert field_name not in ns
+        ns[field_name] = _make_prop(field_type, idx)
+
+
+class AnonProductKind(AnonymousProductMeta, TupleKind, Kind):
     __hash__ = type.__hash__
 
-    def __new__(mcs, name, bases, namespace, cache=True, **kwargs):
-        return super().__new__(mcs, name, bases, namespace, cache, **kwargs)
+    def _from_idx(cls, idx):
+        mcs = type(cls)
 
-    def from_fields(cls, name, fields , cache=True):
-        return super().from_fields(name, fields, cache)
+        try:
+            return mcs._class_cache[cls, idx]
+        except KeyError:
+            pass
 
-    @classmethod
-    def _from_fields(mcs, fields, name, bases, ns, **kwargs):
+        if cls.is_bound:
+            raise TypeError('Type is already bound')
 
-        cls = bases[0]
-
-        # field_name -> tuple index
-        idx_table = dict((k, i) for i,k in enumerate(fields.keys()))
-
-        def _make_prop(field_type, idx):
-            @TypedProperty(field_type)
-            def prop(self):
-                return self[idx]
-
-            @prop.setter
-            def prop(self, value):
-                self[idx] = value
-
-            return prop
+        undirected_idx = OrderedFrozenDict(
+            [(k, v.qualify(Direction.Undirected)) for k, v in idx.items()]
+        )
+        if undirected_idx != idx:
+            bases = [cls[undirected_idx]]
+        else:
+            bases = [cls]
+        bases = tuple(bases)
+        class_name = cls._name_from_idx(idx)
 
+        ns = {'__module__' : cls.__module__}
         # add properties to namespace
         # build properties
-        for field_name, field_type in fields.items():
-            if field_name == "N":
-                # TODO: Make N unreserved
-                raise ValueError("N is a reserved name in Product")
-            assert field_name not in ns
-            idx = idx_table[field_name]
-            ns[field_name] = _make_prop(field_type, idx)
+        _add_properties(ns, idx)
 
-        # this is all really gross but I don't know how to do this cleanly
-        # need to build t so I can call super() in new and init
-        # need to exec to get proper signatures
-        t = TupleKind.__new__(mcs, name, bases, ns, fields=tuple(fields.values()),
-                              **kwargs)
+        t = mcs(class_name, bases, ns, fields=tuple(idx.values()))
+        t._field_table_ = idx
         if t._unbound_base_ is None:
             t._unbound_base_ = cls
-
-        # not strictly necessary could iterative over class dict finding
-        # TypedProperty to reconstruct _field_table_ but that seems bad
-        t._field_table_ = OrderedFrozenDict(fields)
+        mcs._class_cache[cls, idx] = t
+        t._cached_ = True
         return t
 
     def qualify(cls, direction):
@@ -395,12 +517,15 @@ class ProductKind(ProductMeta, TupleKind, Kind):
         for k, v in cls.field_dict.items():
             if not issubclass(new_fields[k], v):
                 base = cls.unbound_t
-        if base.is_bound and all(v == base.field_dict[k] for k, v in
-                                 new_fields.items()):
+        if base.is_bound and all(v is base.field_dict[k]
+                                 for k, v in new_fields.items()):
             return base
 
-        return cls.unbound_t._cache_handler(cls.is_cached, new_fields,
-                                            cls.__name__, (base, ), {})
+        if cls.unbound_t is AnonProduct:
+            return cls.unbound_t._cache_handler(OrderedFrozenDict(new_fields))
+        else:
+            return cls.unbound_t._cache_handler(cls.is_cached, new_fields,
+                                                cls.__name__, (base, ), {})
 
     def flip(cls):
         new_fields = OrderedDict()
@@ -410,20 +535,37 @@ class ProductKind(ProductMeta, TupleKind, Kind):
         for k, v in cls.field_dict.items():
             if not issubclass(new_fields[k], v):
                 base = cls.qualify(Direction.Undirected)
-        return cls.unbound_t._cache_handler(cls.is_cached, new_fields,
-                                            cls.__name__, (base, ), {})
+        if cls.unbound_t is AnonProduct:
+            return cls.unbound_t._cache_handler(OrderedFrozenDict(new_fields))
+        else:
+            return cls.unbound_t._cache_handler(cls.is_cached, new_fields,
+                                                cls.__name__, (base, ), {})
 
     def __eq__(cls, rhs):
-        if not isinstance(rhs, ProductKind):
+        if not isinstance(rhs, AnonProductKind):
             return False
 
         if not cls.is_bound:
             return not rhs.is_bound
 
-        for k, v in cls.field_dict.items():
-            if getattr(rhs, k) != v:
-                return False
+        return cls.field_dict == rhs.field_dict
 
+    def is_wireable(cls, rhs):
+        rhs = magma_type(rhs)
+        if not isinstance(rhs, AnonProductKind) or len(cls.fields) != len(rhs.fields):
+            return False
+        for k, v in cls.field_dict.items():
+            if not k in rhs.field_dict or not v.is_wireable(rhs.field_dict[k]):
+                return False
+        return True
+
+    def is_bindable(cls, rhs):
+        rhs = magma_type(rhs)
+        if not isinstance(rhs, AnonProductKind) or len(cls.fields) != len(rhs.fields):
+            return False
+        for k, v in cls.field_dict.items():
+            if not k in rhs.field_dict or not v.is_bindable(rhs.field_dict[k]):
+                return False
         return True
 
     def __len__(cls):
@@ -438,7 +580,7 @@ class ProductKind(ProductMeta, TupleKind, Kind):
         return s
 
 
-class Product(Tuple, metaclass=ProductKind):
+class AnonProduct(Tuple, metaclass=AnonProductKind):
     @classmethod
     def keys(cls):
         return cls.field_dict.keys()
@@ -458,3 +600,39 @@ class Product(Tuple, metaclass=ProductKind):
             return ts[0].name.tuple
 
         return type(self).flip()(*ts)
+
+
+class ProductKind(ProductMeta, AnonProductKind, Kind):
+    __hash__ = type.__hash__
+
+    def __new__(mcs, name, bases, namespace, cache=True, **kwargs):
+        return super().__new__(mcs, name, bases, namespace, cache, **kwargs)
+
+    def from_fields(cls, name, fields , cache=True):
+        return super().from_fields(name, fields, cache)
+
+    @classmethod
+    def _from_fields(mcs, fields, name, bases, ns, **kwargs):
+
+        cls = bases[0]
+
+        # add properties to namespace
+        # build properties
+        _add_properties(ns, fields)
+
+        # this is all really gross but I don't know how to do this cleanly
+        # need to build t so I can call super() in new and init
+        # need to exec to get proper signatures
+        t = TupleKind.__new__(mcs, name, bases, ns, fields=tuple(fields.values()),
+                              **kwargs)
+        if t._unbound_base_ is None:
+            t._unbound_base_ = cls
+
+        # not strictly necessary could iterative over class dict finding
+        # TypedProperty to reconstruct _field_table_ but that seems bad
+        t._field_table_ = OrderedFrozenDict(fields)
+        return t
+
+
+class Product(AnonProduct, metaclass=ProductKind):
+    pass
